==> TundereBird-0.28.tar.bz2: TundereBird-0.28/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client.Session/ImapSession.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Threading;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Client.Transaction.BuiltIn;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Client.Session {
  // Layers:
  // * Session.ImapSession
  //   Transaction.IImapTransaction/Transaction.ImapTransactionRequest/Transaction.ImapTransactionResult
  //   Protocol.ImapConnection/Protocol.ImapCommand/Protocol.ImapResponse
  //   (IMAP4)

  public partial class ImapSession : IDisposable {
    /*
     * properties
     */
    internal int Id {
      get { return connection.Id; }
    }

    public bool IsDisposed {
      get { return disposed; }
    }

    public bool HandlesReferralAsException {
      get { return handlesReferralAsException; }
      set { handlesReferralAsException = value; }
    }

    public bool HandlesIncapableAsException {
      get { return handlesIncapableAsException; }
      set { handlesIncapableAsException = value; }
    }

    public int SendTimeout {
      get { CheckDisposed(); return connection.SendTimeout; }
      set { CheckDisposed(); RejectTransactionProceeding(); connection.SendTimeout = value; }
    }

    public int ReceiveTimeout {
      get { CheckDisposed(); return connection.ReceiveTimeout; }
      set { CheckDisposed(); RejectTransactionProceeding(); connection.ReceiveTimeout = value; }
    }

    public bool IsSecureConnection {
      get { CheckDisposed(); return connection.IsSecurePortConnection || connection.IsSecureConnection; }
    }

    public int TransactionTimeout {
      get { CheckDisposed(); return transactionTimeout; }
      set
      {
        CheckDisposed();
        RejectTransactionProceeding();

        if (value < -1)
          throw new ArgumentOutOfRangeException("TransactionTimeout", "must be greater than or equals to -1");
        transactionTimeout = value;
      }
    }

    public bool IsTransactionProceeding {
      get
      {
        if (Monitor.TryEnter(transactionLockObject, 0)) {
          Monitor.Exit(transactionLockObject);
          return false;
        }
        else {
          return true;
        }
      }
    }

    /*
     * construction / destruction
     */
    public ImapSession(string host)
      : this(host, ImapDefaultPort.Imap, false, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port)
      : this(host, port, false, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, int transactionTimeout)
      : this(host, port, false, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, false, DefaultTransactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, int port, int transactionTimeout, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, false, transactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, bool handlesReferralAsException)
      : this(host, ImapDefaultPort.Imap, handlesReferralAsException, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, bool handlesReferralAsException, int transactionTimeout)
      : this(host, ImapDefaultPort.Imap, handlesReferralAsException, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException)
      : this(host, port, handlesReferralAsException, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, int transactionTimeout)
      : this(host, port, handlesReferralAsException, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, handlesReferralAsException, DefaultTransactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, int transactionTimeout, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      if (transactionTimeout < -1)
        throw new ArgumentOutOfRangeException("transactionTimeout", "must be greater than or equals to -1");

      this.handlesReferralAsException = handlesReferralAsException;
      this.transactionTimeout = transactionTimeout;

      this.readonlyServerID = serverID.AsReadOnly();
      this.readonlyHierarchyDelimiters = hierarchyDelimiters.AsReadOnly();

      InitializeResponseProcessors();

      Connect(host, port, createAuthenticatedStreamCallback);
    }

    public override string ToString()
    {
      CheckDisposed();

      return string.Format("{{Connection={0}, State={0}}}",
                           connection,
                           state);
    }

    private const int DefaultTransactionTimeout = Timeout.Infinite;

    private bool disposed = false;
    private bool handlesReferralAsException;
    private bool handlesIncapableAsException = false;
    private int transactionTimeout = DefaultTransactionTimeout;
    private object transactionLockObject = new object();

    /*
     * session state properties
     */
    internal ImapConnection Connection {
      get { return connection; }
    }

    public bool IsIdling {
      get { CheckDisposed(); return connection.IsIdling; }
    }

    public ImapSessionState State {
      get { CheckDisposed(); return state; }
    }

    public ImapUpdatedStatus LastUpdatedStatus {
      get { CheckDisposed(); return lastUpdatedStatus; }
      internal set { lastUpdatedStatus = value; }
    }

    public ImapMailbox SelectedMailbox {
      get { CheckDisposed(); return selectedMailbox; }
    }

    public Uri Authority {
      get { CheckDisposed(); return authority; }
    }

    public ImapCapabilityList ServerCapabilities {
      get { CheckDisposed(); return serverCapabilities; }
    }

    public IDictionary<string, string> HierarchyDelimiters {
      get { CheckDisposed(); return readonlyHierarchyDelimiters; }
    }

    public ImapNamespace Namespaces {
      get { CheckDisposed(); return namespaces; }
    }

    public IDictionary<string, string> ServerID {
      get { CheckDisposed(); return readonlyServerID; }
    }

    public string SelectedLanguage {
      get { CheckDisposed(); return selectedLanguage; }
    }

    public ImapCollationAlgorithm ActiveComparator {
      get { CheckDisposed(); return activeComparator; }
    }

    private ImapConnection connection = null;
    private ImapMailboxManager mailboxManager = null;
    private ImapSessionState state = ImapSessionState.NotConnected;
    private ImapUpdatedStatus lastUpdatedStatus = new ImapUpdatedStatus();
    private ImapTransactionResult lastTransactionResult = null;
    private ImapMailbox selectedMailbox = null;
    private Uri authority = null;
    private ImapCapabilityList serverCapabilities;
    private Dictionary<string, string> hierarchyDelimiters = new Dictionary<string, string>();
    private ImapNamespace namespaces = new ImapNamespace();
    private Dictionary<string, string> serverID = new Dictionary<string, string>();
    private string selectedLanguage = null;
    private ImapCollationAlgorithm activeComparator = ImapCollationAlgorithm.Default;

    private IDictionary<string, string> readonlyServerID;
    private IDictionary<string, string> readonlyHierarchyDelimiters;

    /*
     * transaction methods : connect/disconnect
     */
    private void Connect(string host, int port, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      TraceInfo("connecting");

      this.connection = new ImapConnection(host, port, createAuthenticatedStreamCallback);

      TraceInfo("connected");

      SetServerCapabilities(new ImapCapabilityList(new[] {
        ImapCapability.Imap4Rev1 /* required by GreetingTransaction */
      }));

      using (var t = new GreetingTransaction(connection)) {
        try {
          if (ProcessTransaction(t, handlesIncapableAsException).Failed)
            throw new ImapConnectionException("initial greeting failed");
        }
        catch (TimeoutException) {
          throw new ImapConnectionException("initial greeting failed");
        }

        if (t.Result.Code == ImapTransactionResultCode.PreAuth) {
          UpdateAuthority(null, null);
          TransitStateTo(ImapSessionState.Authenticated);
        }
        else {
          TransitStateTo(ImapSessionState.NotAuthenticated);
        }

        if (!t.Result.ReceivedResponses.Exists(delegate(ImapResponse response) {
          return (response is ImapStatusResponse) &&
                 (response as ImapStatusResponse).ResponseText.Code == ImapResponseCode.Capability;
        })) {
          // clear server capabilities
          SetServerCapabilities(new ImapCapabilityList());
        }
      }
    }

    /// <summary>disconnects session</summary>
    /// <remarks>this method sends CLOSE and LOGOUT before disconnect the session.</remarks>
    public void Disconnect()
    {
      Disconnect(true);
    }

    /// <summary>disconnects session</summary>
    public void Disconnect(bool logout)
    {
      CheckDisposed();

      if (logout) {
        if (state == ImapSessionState.Selected)
          Close();
        if (state == ImapSessionState.Authenticated)
          Logout();
      }

      (this as IDisposable).Dispose();
    }

    private void CloseConnection()
    {
      if (connection == null)
        return;

      try {
        TraceInfo("disconnecting");

        connection.Close();

        TraceInfo("disconnected");
      }
      finally {
        TransitStateTo(ImapSessionState.NotConnected);

        connection = null;
      }
    }

    void IDisposable.Dispose()
    {
      if (disposed)
        return;

      CloseConnection();

      disposed = true;
    }

    /*
     * transaction methods : generic use, exntension, etc.
     */

    /// <summary>sends generic/undefined command</summary>
    public ImapTransactionResult GenericCommand(string command, params ImapString[] arguments)
    {
      ImapDataResponse[] discard;

      return GenericCommand(command, out discard, arguments);
    }

    /// <summary>sends generic/undefined command</summary>
    public ImapTransactionResult GenericCommand(string command, out ImapDataResponse[] dataResponses, params ImapString[] arguments)
    {
      RejectNonConnectedState();

      if (command == null)
        throw new ArgumentNullException("command");
      else if (command.Length == 0)
        throw new ArgumentException("command must be a non-empty string", "command");

      dataResponses = null;

      using (var t = new GenericCommandTransaction(connection, command)) {
        if (arguments != null && 0 < arguments.Length)
          t.Request.Arguments["arguments"] = new ImapStringList(arguments);

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          dataResponses = t.Result.Value;

        return t.Result;
      }
    }

    /*
     * methods for internal state management
     */
    internal void SetServerCapabilities(ImapCapabilityList newCapabilities)
    {
      serverCapabilities = new ImapCapabilityList(true, newCapabilities);

#if false
      defaultLiteralSynchronizationMode = serverCapabilities.Has(ImapCapability.LiteralNonSync)
        ? ImapLiteralOptions.NonSynchronizing
        : ImapLiteralOptions.Synchronizing;
#endif
    }

    private void TransitStateTo(ImapSessionState newState)
    {
      if (newState == state)
        return;

      switch (newState) {
        case ImapSessionState.NotConnected:
          authority = null;
          selectedMailbox = null;
          if (mailboxManager != null) {
            mailboxManager.DetachFromSession();
            mailboxManager = null;
          }
          TraceInfo("now in non-connected state");
          break;

        case ImapSessionState.NotAuthenticated:
          UpdateAuthority(null, null);
          if (state == ImapSessionState.Selected || state == ImapSessionState.Authenticated) {
            mailboxManager.DetachFromSession();
            mailboxManager = null;
            TraceInfo("logged out");
          }
          TraceInfo("now in non-authenticated state");
          break;

        case ImapSessionState.Authenticated:
          TraceInfo("now in authenticated state");
          if (state == ImapSessionState.NotAuthenticated || state == ImapSessionState.NotConnected) {
            mailboxManager = new ImapMailboxManager(this);
            TraceInfo("logged in");
          }
          break;

        case ImapSessionState.Selected:
          TraceInfo("now in selected state (selected '{0}')", selectedMailbox.Name);
          break;
      }

      state = newState;
    }

    /*
     * methods for checking and validating session status
     */
    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private void RejectNonConnectedState()
    {
      CheckDisposed();

      if (state == ImapSessionState.NotConnected)
        throw new ImapProtocolViolationException("session has been disconnected");
    }

    private void RejectNonAuthenticatedState()
    {
      RejectNonConnectedState();

      if (state == ImapSessionState.NotAuthenticated)
        throw new ImapProtocolViolationException("session is not in authenticated state");
    }

    private void RejectNonSelectedState()
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.Selected)
        throw new ImapProtocolViolationException("session is not in selected state");
    }

    private void RejectIdling()
    {
      if (IsIdling)
        throw new ImapProtocolViolationException("session is idling now");
    }

    private void ValidateMailboxRelationship(ImapMailbox mailbox)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");

      if (mailbox.Session != this)
        throw new ArgumentException("mailbox is not attached to current session", "mailbox");
    }

    private void CheckServerCapability(IImapExtension extension)
    {
      CheckServerCapability(extension.RequiredCapability);
    }

    private void CheckServerCapability(IImapMultipleExtension extensions)
    {
      foreach (var capability in extensions.RequiredCapabilities) {
        CheckServerCapability(capability);
      }
    }

    private void CheckServerCapability(ImapCapability capability)
    {
      if (capability != null && !serverCapabilities.Has(capability))
        throw new ImapIncapableException(capability);
    }

    private void RejectInvalidMailboxNameArgument(string mailboxName)
    {
      if (mailboxName == null)
        throw new ArgumentException("mailboxName");
      else if (mailboxName.Length == 0)
        throw new ArgumentException("invalid mailbox name", "mailboxName");
    }

    private void RejectTransactionProceeding()
    {
      if (IsTransactionProceeding)
        throw new InvalidOperationException("transaction proceeding");
    }

    /*
     * tracing
     */
    [System.Diagnostics.Conditional("TRACE")]
    private void TraceInfo(string format, params object[] arguments)
    {
      Trace.Log(this, format, arguments);
    }
  }
}

==> TundereBird-0.29.tar.bz2: TundereBird-0.29/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client.Session/ImapSession.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Threading;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Client.Transaction.BuiltIn;

namespace Smdn.Net.Imap4.Client.Session {
  // Layers:
  // * Session.ImapSession
  //   Transaction.IImapTransaction/Transaction.ImapTransactionRequest/Transaction.ImapTransactionResult
  //   Protocol.ImapConnection/Protocol.ImapCommand/Protocol.ImapResponse
  //   (IMAP4)

  public partial class ImapSession : IDisposable {
    /*
     * properties
     */
    internal int? Id {
      get { return (connection == null) ? (int?)null : connection.Id; }
    }

    public bool IsDisposed {
      get { return disposed; }
    }

    public bool HandlesReferralAsException {
      get { return handlesReferralAsException; }
      set { handlesReferralAsException = value; }
    }

    public bool HandlesIncapableAsException {
      get { return handlesIncapableAsException; }
      set { handlesIncapableAsException = value; }
    }

    public int SendTimeout {
      get { CheckDisposed(); return connection.SendTimeout; }
      set { CheckDisposed(); RejectTransactionProceeding(); connection.SendTimeout = value; }
    }

    public int ReceiveTimeout {
      get { CheckDisposed(); return connection.ReceiveTimeout; }
      set { CheckDisposed(); RejectTransactionProceeding(); connection.ReceiveTimeout = value; }
    }

    public bool IsSecureConnection {
      get { CheckDisposed(); return connection.IsSecurePortConnection || connection.IsSecureConnection; }
    }

    public int TransactionTimeout {
      get { CheckDisposed(); return transactionTimeout; }
      set
      {
        CheckDisposed();
        RejectTransactionProceeding();

        if (value < -1)
          throw new ArgumentOutOfRangeException("TransactionTimeout", "must be greater than or equals to -1");
        transactionTimeout = value;
      }
    }

    public bool IsTransactionProceeding {
      get
      {
        if (Monitor.TryEnter(transactionLockObject, 0)) {
          Monitor.Exit(transactionLockObject);
          return false;
        }
        else {
          return true;
        }
      }
    }

    /*
     * construction / destruction
     */
    public ImapSession(string host)
      : this(host, ImapDefaultPort.Imap, false, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port)
      : this(host, port, false, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, int transactionTimeout)
      : this(host, port, false, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, false, DefaultTransactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, int port, int transactionTimeout, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, false, transactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, bool handlesReferralAsException)
      : this(host, ImapDefaultPort.Imap, handlesReferralAsException, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, bool handlesReferralAsException, int transactionTimeout)
      : this(host, ImapDefaultPort.Imap, handlesReferralAsException, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException)
      : this(host, port, handlesReferralAsException, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, int transactionTimeout)
      : this(host, port, handlesReferralAsException, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, handlesReferralAsException, DefaultTransactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, int transactionTimeout, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      if (transactionTimeout < -1)
        throw new ArgumentOutOfRangeException("transactionTimeout", "must be greater than or equals to -1");

      this.handlesReferralAsException = handlesReferralAsException;
      this.transactionTimeout = transactionTimeout;

      this.readonlyServerID = serverID.AsReadOnly();
      this.readonlyHierarchyDelimiters = hierarchyDelimiters.AsReadOnly();

      InitializeResponseProcessors();

      Connect(host, port, createAuthenticatedStreamCallback);
    }

    public override string ToString()
    {
      CheckDisposed();

      return string.Format("{{Connection={0}, State={0}}}",
                           connection,
                           state);
    }

    private const int DefaultTransactionTimeout = Timeout.Infinite;

    private bool disposed = false;
    private bool handlesReferralAsException;
    private bool handlesIncapableAsException = false;
    private int transactionTimeout = DefaultTransactionTimeout;
    private object transactionLockObject = new object();

    /*
     * session state properties
     */
    public IConnectionInfo ConnectionInfo {
      get { CheckDisposed(); return connection; }
    }

    public bool IsIdling {
      get { CheckDisposed(); return connection.IsIdling; }
    }

    public ImapSessionState State {
      get { CheckDisposed(); return state; }
    }

    public ImapUpdatedStatus LastUpdatedStatus {
      get { CheckDisposed(); return lastUpdatedStatus; }
      internal set { lastUpdatedStatus = value; }
    }

    public ImapMailbox SelectedMailbox {
      get { CheckDisposed(); return selectedMailbox; }
    }

    public Uri Authority {
      get { CheckDisposed(); return authority == null ? null : authority.Uri; }
    }

    internal ImapUriBuilder AuthorityBuilder {
      get { return authority; }
    }

    public ImapCapabilityList ServerCapabilities {
      get { CheckDisposed(); return serverCapabilities; }
    }

    public IDictionary<string, string> HierarchyDelimiters {
      get { CheckDisposed(); return readonlyHierarchyDelimiters; }
    }

    public ImapNamespace Namespaces {
      get { CheckDisposed(); return namespaces; }
    }

    public IDictionary<string, string> ServerID {
      get { CheckDisposed(); return readonlyServerID; }
    }

    public string SelectedLanguage {
      get { CheckDisposed(); return selectedLanguage; }
    }

    public ImapCollationAlgorithm ActiveComparator {
      get { CheckDisposed(); return activeComparator; }
    }

    private ImapConnection connection = null;
    private ImapMailboxManager mailboxManager = null;
    private ImapSessionState state = ImapSessionState.NotConnected;
    private ImapUpdatedStatus lastUpdatedStatus = new ImapUpdatedStatus();
    private ImapTransactionResult lastTransactionResult = null;
    private ImapMailbox selectedMailbox = null;
    private ImapUriBuilder authority = new ImapUriBuilder();
    private ImapCapabilityList serverCapabilities;
    private Dictionary<string, string> hierarchyDelimiters = new Dictionary<string, string>();
    private ImapNamespace namespaces = new ImapNamespace();
    private Dictionary<string, string> serverID = new Dictionary<string, string>();
    private string selectedLanguage = null;
    private ImapCollationAlgorithm activeComparator = ImapCollationAlgorithm.Default;

    private IDictionary<string, string> readonlyServerID;
    private IDictionary<string, string> readonlyHierarchyDelimiters;

    /*
     * transaction methods : connect/disconnect
     */
    private void Connect(string host, int port, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      TraceInfo("connecting");

      this.connection = new ImapConnection(host, port, createAuthenticatedStreamCallback);

      TraceInfo("connected");

      SetServerCapabilities(new ImapCapabilityList(new[] {
        ImapCapability.Imap4Rev1 /* required by GreetingTransaction */
      }));

      using (var t = new GreetingTransaction(connection)) {
        try {
          if (ProcessTransaction(t, handlesIncapableAsException).Failed)
            throw new ImapConnectionException("initial greeting failed");
        }
        catch (TimeoutException) {
          throw new ImapConnectionException("initial greeting failed");
        }

        if (t.Result.Code == ImapTransactionResultCode.PreAuth) {
          UpdateAuthority(null, null);
          TransitStateTo(ImapSessionState.Authenticated);
        }
        else {
          TransitStateTo(ImapSessionState.NotAuthenticated);
        }

        if (!t.Result.ReceivedResponses.Exists(ImapCapabilityResponseProcessor.IsCapabilityResponse))
          // clear server capabilities
          SetServerCapabilities(new ImapCapabilityList());
      }
    }

    /// <summary>disconnects session</summary>
    /// <remarks>this method sends CLOSE and LOGOUT before disconnect the session.</remarks>
    public void Disconnect()
    {
      Disconnect(true);
    }

    /// <summary>disconnects session</summary>
    public void Disconnect(bool logout)
    {
      CheckDisposed();

      if (logout) {
        if (state == ImapSessionState.Selected)
          Close();
        if (state == ImapSessionState.Authenticated)
          Logout();
      }

      (this as IDisposable).Dispose();
    }

    private void CloseConnection()
    {
      if (connection == null)
        return;

      try {
        TraceInfo("disconnecting");

        connection.Close();

        TraceInfo("disconnected");
      }
      finally {
        TransitStateTo(ImapSessionState.NotConnected);

        connection = null;
      }
    }

    void IDisposable.Dispose()
    {
      if (disposed)
        return;

      CloseConnection();

      disposed = true;
    }

    /*
     * transaction methods : generic use, exntension, etc.
     */

    /// <summary>sends generic/undefined command</summary>
    public ImapTransactionResult GenericCommand(string command, params ImapString[] arguments)
    {
      ImapDataResponse[] discard;

      return GenericCommand(command, out discard, arguments);
    }

    /// <summary>sends generic/undefined command</summary>
    public ImapTransactionResult GenericCommand(string command, out ImapDataResponse[] dataResponses, params ImapString[] arguments)
    {
      RejectNonConnectedState();

      if (command == null)
        throw new ArgumentNullException("command");
      else if (command.Length == 0)
        throw new ArgumentException("command must be a non-empty string", "command");

      dataResponses = null;

      using (var t = new GenericCommandTransaction(connection, command)) {
        if (arguments != null && 0 < arguments.Length)
          t.Request.Arguments["arguments"] = new ImapStringList(arguments);

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          dataResponses = t.Result.Value;

        return t.Result;
      }
    }

    /*
     * methods for internal state management
     */
    internal void SetServerCapabilities(ImapCapabilityList newCapabilities)
    {
      serverCapabilities = new ImapCapabilityList(true, newCapabilities);

#if false
      defaultLiteralSynchronizationMode = serverCapabilities.Has(ImapCapability.LiteralNonSync)
        ? ImapLiteralOptions.NonSynchronizing
        : ImapLiteralOptions.Synchronizing;
#endif
    }

    private void TransitStateTo(ImapSessionState newState)
    {
      if (newState == state)
        return;

      switch (newState) {
        case ImapSessionState.NotConnected:
          authority = null;
          selectedMailbox = null;
          if (mailboxManager != null) {
            mailboxManager.DetachFromSession();
            mailboxManager = null;
          }
          TraceInfo("now in non-connected state");
          break;

        case ImapSessionState.NotAuthenticated:
          UpdateAuthority(null, null);
          if (state == ImapSessionState.Selected || state == ImapSessionState.Authenticated) {
            mailboxManager.DetachFromSession();
            mailboxManager = null;
            TraceInfo("logged out");
          }
          TraceInfo("now in non-authenticated state");
          break;

        case ImapSessionState.Authenticated:
          TraceInfo("now in authenticated state");
          if (state == ImapSessionState.NotAuthenticated || state == ImapSessionState.NotConnected) {
            mailboxManager = new ImapMailboxManager(this);
            TraceInfo("logged in");
          }
          break;

        case ImapSessionState.Selected:
          TraceInfo("now in selected state (selected '{0}')", selectedMailbox.Name);
          break;
      }

      state = newState;
    }

    /*
     * methods for checking and validating session status
     */
    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private void RejectNonConnectedState()
    {
      CheckDisposed();

      if (state == ImapSessionState.NotConnected)
        throw new ImapProtocolViolationException("session has been disconnected");
    }

    private void RejectNonAuthenticatedState()
    {
      RejectNonConnectedState();

      if (state == ImapSessionState.NotAuthenticated)
        throw new ImapProtocolViolationException("session is not in authenticated state");
    }

    private void RejectNonSelectedState()
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.Selected)
        throw new ImapProtocolViolationException("session is not in selected state");
    }

    private void RejectIdling()
    {
      if (IsIdling)
        throw new ImapProtocolViolationException("session is idling now");
    }

    private void ValidateMailboxRelationship(ImapMailbox mailbox)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");

      if (mailbox.Session != this)
        throw new ArgumentException("mailbox is not attached to current session", "mailbox");
    }

    private void CheckServerCapability(IImapExtension extension)
    {
      CheckServerCapability(extension.RequiredCapability);
    }

    private void CheckServerCapability(IImapMultipleExtension extensions)
    {
      foreach (var capability in extensions.RequiredCapabilities) {
        CheckServerCapability(capability);
      }
    }

    private void CheckServerCapability(ImapCapability capability)
    {
      if (capability != null && !serverCapabilities.Has(capability))
        throw new ImapIncapableException(capability);
    }

    private void RejectInvalidMailboxNameArgument(string mailboxName)
    {
      if (mailboxName == null)
        throw new ArgumentException("mailboxName");
      else if (mailboxName.Length == 0)
        throw new ArgumentException("invalid mailbox name", "mailboxName");
    }

    private void RejectTransactionProceeding()
    {
      if (IsTransactionProceeding)
        throw new InvalidOperationException("transaction proceeding");
    }

    /*
     * tracing
     */
    [System.Diagnostics.Conditional("TRACE")]
    private void TraceInfo(string format, params object[] arguments)
    {
      Trace.Log(this, format, arguments);
    }
  }
}
